@@ -2,20 +2,26 @@ use crate::build_options::{extract_cargo_metadata_args, CargoOptions};
 use crate::{CargoToml, Metadata21, PyProjectToml};
 use anyhow::{bail, format_err, Context, Result};
 use cargo_metadata::{Metadata, MetadataCommand};
-use fs_err as fs;
+use normpath::PathExt as _;
+use std::collections::HashSet;
 use std::env;
 use std::io;
 use std::path::{Path, PathBuf};
+use tracing::debug;
 
 const PYPROJECT_TOML: &str = "pyproject.toml";
 
 /// Whether this project is pure rust or rust mixed with python and whether it has wheel data
 #[derive(Clone, Debug, PartialEq, Eq)]
 pub struct ProjectLayout {
+    /// Contains the absolute path to the python source directory
+    pub python_dir: PathBuf,
     /// Contains the canonicalized (i.e. absolute) path to the python part of the project
     /// If none, we have a rust crate compiled into a shared library with only some glue python for cffi
     /// If some, we have a python package that is extended by a native rust module.
     pub python_module: Option<PathBuf>,
+    /// Python packages to include
+    pub python_packages: Vec<String>,
     /// Contains the canonicalized (i.e. absolute) path to the rust part of the project
     pub rust_module: PathBuf,
     /// Rust extension name
@@ -63,8 +69,23 @@ impl ProjectResolver {
                 manifest_file.display()
             );
         }
+        // Sanity checks in debug build
+        debug_assert!(
+            manifest_file.is_absolute(),
+            "manifest_file {} is not absolute",
+            manifest_file.display()
+        );
+        debug_assert!(
+            pyproject_file.is_absolute(),
+            "pyproject_file {} is not absolute",
+            pyproject_file.display()
+        );
+
+        // Set Cargo manifest path
+        cargo_options.manifest_path = Some(manifest_file.clone());
+
         let cargo_toml = CargoToml::from_path(&manifest_file)?;
-        cargo_toml.warn_deprecated_python_metadata();
+        cargo_toml.warn_removed_python_metadata();
 
         let manifest_dir = manifest_file.parent().unwrap();
         let pyproject_toml: Option<PyProjectToml> = if pyproject_file.is_file() {
@@ -88,11 +109,11 @@ impl ProjectResolver {
         let cargo_metadata = Self::resolve_cargo_metadata(&manifest_file, &cargo_options)?;
 
         let mut metadata21 =
-            Metadata21::from_cargo_toml(&cargo_toml, &manifest_dir, &cargo_metadata)
+            Metadata21::from_cargo_toml(&cargo_toml, manifest_dir, &cargo_metadata)
                 .context("Failed to parse Cargo.toml into python metadata")?;
         if let Some(pyproject) = pyproject {
             let pyproject_dir = pyproject_file.parent().unwrap();
-            metadata21.merge_pyproject_toml(&pyproject_dir, pyproject)?;
+            metadata21.merge_pyproject_toml(pyproject_dir, pyproject)?;
         }
         let extra_metadata = cargo_toml.remaining_core_metadata();
 
@@ -119,11 +140,43 @@ impl ProjectResolver {
         } else {
             manifest_dir
         };
+        let python_packages = pyproject
+            .and_then(|x| x.python_packages())
+            .unwrap_or_default()
+            .to_vec();
         let py_root = match pyproject.and_then(|x| x.python_source()) {
-            Some(py_src) => py_src.to_path_buf(),
+            Some(py_src) => project_root.join(py_src),
             None => match extra_metadata.python_source.as_ref() {
-                Some(py_src) => manifest_dir.join(py_src),
-                None => project_root.to_path_buf(),
+                Some(py_src) => {
+                    println!("⚠️ Warning: specify python-source in Cargo.toml is deprecated, use python-source in [tool.maturin] section in pyproject.toml instead");
+                    manifest_dir.join(py_src)
+                }
+                None => match pyproject.and_then(|x| x.project_name()) {
+                    Some(project_name) => {
+                        // Detect src layout
+                        let rust_cargo_toml_found =
+                            project_root.join("rust").join("Cargo.toml").is_file();
+                        let import_name = project_name.replace('-', "_");
+                        let mut package_init = HashSet::new();
+                        package_init.insert(
+                            project_root
+                                .join("src")
+                                .join(import_name)
+                                .join("__init__.py"),
+                        );
+                        for package in &python_packages {
+                            package_init
+                                .insert(project_root.join("src").join(package).join("__init__.py"));
+                        }
+                        let python_src_found = package_init.iter().any(|x| x.is_file());
+                        if rust_cargo_toml_found && python_src_found {
+                            project_root.join("src")
+                        } else {
+                            project_root.to_path_buf()
+                        }
+                    }
+                    None => project_root.to_path_buf(),
+                },
             },
         };
         let data = match pyproject.and_then(|x| x.data()) {
@@ -143,7 +196,8 @@ impl ProjectResolver {
                 }
             }),
         };
-        let project_layout = ProjectLayout::determine(project_root, extension_name, py_root, data)?;
+        let project_layout =
+            ProjectLayout::determine(project_root, extension_name, py_root, python_packages, data)?;
         Ok(Self {
             project_layout,
             cargo_toml_path: manifest_file,
@@ -165,43 +219,93 @@ impl ProjectResolver {
     ) -> Result<(PathBuf, PathBuf)> {
         // use command line argument if specified
         if let Some(path) = cargo_manifest_path {
+            let path = path.normalize()?.into_path_buf();
+            debug!(
+                "Using cargo manifest path from command line argument: {:?}",
+                path
+            );
             let workspace_root = Self::resolve_cargo_metadata(&path, cargo_options)?.workspace_root;
-            for parent in fs::canonicalize(&path)?.ancestors().skip(1) {
-                if !parent.starts_with(&workspace_root) {
+            let workspace_parent = workspace_root.parent().unwrap_or(&workspace_root);
+            for parent in path.ancestors().skip(1) {
+                // Allow looking outside to the parent directory of Cargo workspace root
+                if !dunce::simplified(parent).starts_with(workspace_parent) {
                     break;
                 }
                 let pyproject_file = parent.join(PYPROJECT_TOML);
                 if pyproject_file.is_file() {
-                    // Don't return canonicalized manifest path
-                    // cargo doesn't handle them well.
-                    // See https://github.com/rust-lang/cargo/issues/9770
+                    debug!("Found pyproject.toml at {:?}", pyproject_file);
                     return Ok((path, pyproject_file));
                 }
             }
-            return Ok((path.clone(), path.parent().unwrap().join(PYPROJECT_TOML)));
+            let pyproject_file = path.parent().unwrap().join(PYPROJECT_TOML);
+            debug!("Trying pyproject.toml at {:?}", pyproject_file);
+            return Ok((path, pyproject_file));
         }
         // check `manifest-path` option in pyproject.toml
-        let current_dir = fs::canonicalize(
-            env::current_dir().context("Failed to detect current directory ಠ_ಠ")?,
-        )?;
+        let current_dir = env::current_dir()
+            .context("Failed to detect current directory ಠ_ಠ")?
+            .normalize()?
+            .into_path_buf();
         let pyproject_file = current_dir.join(PYPROJECT_TOML);
         if pyproject_file.is_file() {
+            debug!(
+                "Found pyproject.toml in working directory at {:?}",
+                pyproject_file
+            );
             let pyproject =
                 PyProjectToml::new(&pyproject_file).context("pyproject.toml is invalid")?;
             if let Some(path) = pyproject.manifest_path() {
-                // pyproject.toml must be placed at top directory
-                let manifest_dir =
-                    fs::canonicalize(path.parent().context("missing parent directory")?)?;
-                if !manifest_dir.starts_with(&current_dir) {
-                    bail!("Cargo.toml can not be placed outside of the directory containing pyproject.toml");
+                debug!("Using cargo manifest path from pyproject.toml {:?}", path);
+                return Ok((path.normalize()?.into_path_buf(), pyproject_file));
+            } else {
+                // Detect src layout:
+                //
+                // my-project
+                // ├── README.md
+                // ├── pyproject.toml
+                // ├── src
+                // │   └── my_project
+                // │       ├── __init__.py
+                // │       └── bar.py
+                // └── rust
+                //     ├── Cargo.toml
+                //     └── src
+                //         └── lib.rs
+                let path = current_dir.join("rust").join("Cargo.toml");
+                if path.is_file() {
+                    debug!("Python first src-layout detected");
+                    if pyproject.python_source().is_some() {
+                        // python source directory is specified in pyproject.toml
+                        return Ok((path, pyproject_file));
+                    } else if let Some(project_name) = pyproject.project_name() {
+                        // Check if python source directory in `src/<project_name>`
+                        let import_name = project_name.replace('-', "_");
+                        let mut package_init = HashSet::new();
+                        package_init.insert(
+                            current_dir
+                                .join("src")
+                                .join(import_name)
+                                .join("__init__.py"),
+                        );
+                        for package in pyproject.python_packages().unwrap_or_default() {
+                            package_init
+                                .insert(current_dir.join("src").join(package).join("__init__.py"));
+                        }
+                        if package_init.iter().any(|x| x.is_file()) {
+                            return Ok((path, pyproject_file));
+                        }
+                    }
                 }
-                return Ok((path.to_path_buf(), pyproject_file));
             }
         }
         // check Cargo.toml in current directory
-        let path = PathBuf::from("Cargo.toml");
+        let path = current_dir.join("Cargo.toml");
         if path.exists() {
-            Ok((path, PathBuf::from(PYPROJECT_TOML)))
+            debug!(
+                "Using cargo manifest path from working directory: {:?}",
+                path
+            );
+            Ok((path, current_dir.join(PYPROJECT_TOML)))
         } else {
             Err(format_err!(
                 "Can't find {} (in {})",
@@ -215,9 +319,11 @@ impl ProjectResolver {
         manifest_path: &Path,
         cargo_options: &CargoOptions,
     ) -> Result<Metadata> {
+        debug!("Resolving cargo metadata from {:?}", manifest_path);
         let cargo_metadata_extra_args = extract_cargo_metadata_args(cargo_options)?;
         let result = MetadataCommand::new()
             .manifest_path(manifest_path)
+            .verbose(true)
             .other_options(cargo_metadata_extra_args)
             .exec();
 
@@ -243,6 +349,7 @@ impl ProjectLayout {
         project_root: impl AsRef<Path>,
         module_name: &str,
         python_root: PathBuf,
+        python_packages: Vec<String>,
         data: Option<PathBuf>,
     ) -> Result<ProjectLayout> {
         // A dot in the module name means the extension module goes into the module folder specified by the path
@@ -263,6 +370,14 @@ impl ProjectLayout {
                 module_name.to_string(),
             )
         };
+        debug!(
+            project_root = %project_root.display(),
+            python_dir = %python_root.display(),
+            rust_module = %rust_module.display(),
+            python_module = %python_module.display(),
+            extension_name = %extension_name,
+            "Project layout resolved"
+        );
 
         let data = if let Some(data) = data {
             if !data.is_dir() {
@@ -285,6 +400,8 @@ impl ProjectLayout {
             println!("🍹 Building a mixed python/rust project");
 
             Ok(ProjectLayout {
+                python_dir: python_root,
+                python_packages,
                 python_module: Some(python_module),
                 rust_module,
                 extension_name,
@@ -292,6 +409,8 @@ impl ProjectLayout {
             })
         } else {
             Ok(ProjectLayout {
+                python_dir: python_root,
+                python_packages,
                 python_module: None,
                 rust_module: project_root.to_path_buf(),
                 extension_name,
